// Code specific to the client in the client-server mode.

use mgen::updater::Updater;
use mgen::{HandshakeRef, MessageHeader, SerializedMessage};
use rand_xoshiro::{rand_core::SeedableRng, Xoshiro256PlusPlus};
use serde::Deserialize;
use std::result::Result;
use tokio::io::AsyncWriteExt;
use tokio::net::tcp::{OwnedReadHalf, OwnedWriteHalf};
use tokio::sync::mpsc;
use tokio::task;
use tokio::time::Duration;
use tokio_socks::tcp::Socks5Stream;

mod messenger;

use crate::messenger::dists::{ConfigDistributions, Distributions};
use crate::messenger::error::{FatalError, MessengerError};
use crate::messenger::state::{
    manage_active_conversation, manage_idle_conversation, StateMachine, StateToWriter,
};

/// Type for sending messages from the reader thread to the state thread.
type ReaderToState = mpsc::UnboundedSender<MessageHeader>;
/// Type of messages sent to the writer thread.
type MessageHolder = Box<SerializedMessage>;
/// Type for getting messages from the state thread in the writer thread.
type WriterFromState = mpsc::UnboundedReceiver<MessageHolder>;
/// Type for sending the updated read half of the socket.
type ReadSocketUpdaterIn = Updater<OwnedReadHalf>;
/// Type for getting the updated read half of the socket.
type ReadSocketUpdaterOut = Updater<OwnedReadHalf>;
/// Type for sending the updated write half of the socket.
type WriteSocketUpdaterIn = Updater<OwnedWriteHalf>;
/// Type for getting the updated write half of the socket.
type WriteSocketUpdaterOut = Updater<OwnedWriteHalf>;
/// Type for sending errors to other threads.
type ErrorChannelIn = mpsc::UnboundedSender<MessengerError>;
/// Type for getting errors from other threads.
type ErrorChannelOut = mpsc::UnboundedReceiver<MessengerError>;

/// The thread responsible for getting incoming messages,
/// checking for any network errors while doing so,
/// and giving messages to the state thread.
async fn reader(
    message_channel: ReaderToState,
    mut socket_updater: ReadSocketUpdaterOut,
    error_channel: ErrorChannelIn,
) {
    loop {
        let mut stream = socket_updater.recv().await;
        loop {
            let msg = match mgen::get_message(&mut stream).await {
                Ok(msg) => msg,
                Err(e) => {
                    error_channel.send(e.into()).expect("Error channel closed");
                    break;
                }
            };

            if msg.body != mgen::MessageBody::Receipt {
                message_channel
                    .send(msg)
                    .expect("Reader message channel closed");
            }
        }
    }
}

/// The thread responsible for sending messages from the state thread,
/// and checking for any network errors while doing so.
async fn writer(
    mut message_channel: WriterFromState,
    mut socket_updater: WriteSocketUpdaterOut,
    error_channel: ErrorChannelIn,
) {
    loop {
        let mut stream = socket_updater.recv().await;
        loop {
            let msg = message_channel
                .recv()
                .await
                .expect("Writer message channel closed");
            if let Err(e) = msg.write_all_to(&mut stream).await {
                error_channel.send(e.into()).expect("Error channel closed");
                break;
            }
        }
    }
}

/// Parameters used in establishing the connection through the socks proxy.
/// (Members may be useful elsewhere as well, but that's the primary purpose.)
struct SocksParams {
    socks: String,
    server: String,
    user: String,
    group: String,
}

/// The thread responsible for (re-)establishing connections to the server,
/// and determining how to handle errors this or other threads receive.
async fn socket_updater(
    str_params: SocksParams,
    bootstrap: f64,
    retry: f64,
    mut error_channel: ErrorChannelOut,
    reader_channel: ReadSocketUpdaterIn,
    writer_channel: WriteSocketUpdaterIn,
) -> FatalError {
    let mut first_run: bool = true;
    let retry = Duration::from_secs_f64(retry);
    loop {
        let socks_connection: Result<Socks5Stream<_>, MessengerError> =
            Socks5Stream::connect_with_password(
                str_params.socks.as_str(),
                str_params.server.as_str(),
                &str_params.user,
                &str_params.group,
            )
            .await
            .map_err(|e| e.into());
        let mut stream = match socks_connection {
            Ok(stream) => stream,
            Err(MessengerError::Recoverable(_)) => {
                tokio::time::sleep(retry).await;
                continue;
            }
            Err(MessengerError::Fatal(e)) => return e,
        };

        let handshake = HandshakeRef {
            sender: &str_params.user,
            group: &str_params.group,
        };

        if stream.write_all(&handshake.serialize()).await.is_err() {
            continue;
        }

        if first_run {
            tokio::time::sleep(Duration::from_secs_f64(bootstrap)).await;
            first_run = false;
        }

        let (rd, wr) = stream.into_inner().into_split();
        reader_channel.send(rd);
        writer_channel.send(wr);

        let res = error_channel.recv().await.expect("Error channel closed");
        if let MessengerError::Fatal(e) = res {
            return e;
        }
    }
}

/// The thread responsible for handling the conversation state
/// (i.e., whether the user is active or idle, and when to send messages).
/// Spawns all other threads for this conversation.
async fn manage_conversation(config: Config) -> Result<(), MessengerError> {
    let mut rng = Xoshiro256PlusPlus::from_entropy();
    let distributions: Distributions = config.distributions.try_into()?;

    let str_params = SocksParams {
        socks: config.socks,
        server: config.server,
        user: config.user.clone(),
        group: config.group.clone(),
    };

    let mut state_machine = StateMachine::start(distributions, &mut rng);
    let recipients: Vec<&str> = config.recipients.iter().map(String::as_str).collect();

    let (reader_to_state, mut state_from_reader) = mpsc::unbounded_channel();
    let (state_to_writer, writer_from_state) = mpsc::unbounded_channel();
    let read_socket_updater_in = Updater::new();
    let read_socket_updater_out = read_socket_updater_in.clone();
    let write_socket_updater_in = Updater::new();
    let write_socket_updater_out = write_socket_updater_in.clone();
    let (errs_in, errs_out) = mpsc::unbounded_channel();
    tokio::spawn(reader(
        reader_to_state,
        read_socket_updater_out,
        errs_in.clone(),
    ));
    tokio::spawn(writer(writer_from_state, write_socket_updater_out, errs_in));
    tokio::spawn(socket_updater(
        str_params,
        config.bootstrap,
        config.retry,
        errs_out,
        read_socket_updater_in,
        write_socket_updater_in,
    ));

    let mut state_to_writer = StateToWriter {
        channel: state_to_writer,
    };
    loop {
        state_machine = match state_machine {
            StateMachine::Idle(conversation) => {
                manage_idle_conversation(
                    conversation,
                    &mut state_from_reader,
                    &mut state_to_writer,
                    &config.user,
                    &config.group,
                    recipients.clone(),
                    &mut rng,
                )
                .await
            }
            StateMachine::Active(conversation) => {
                manage_active_conversation(
                    conversation,
                    &mut state_from_reader,
                    &mut state_to_writer,
                    &config.user,
                    &config.group,
                    recipients.clone(),
                    &mut rng,
                )
                .await
            }
        };
    }
}

#[derive(Debug, Deserialize)]
struct Config {
    user: String,
    group: String,
    recipients: Vec<String>,
    socks: String,
    server: String,
    bootstrap: f64,
    retry: f64,
    distributions: ConfigDistributions,
}

#[tokio::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    let mut args = std::env::args();
    let _ = args.next();
    let mut handles = vec![];
    for config_file in args {
        let toml_s = std::fs::read_to_string(config_file)?;
        let config = toml::from_str(&toml_s)?;
        let handle: task::JoinHandle<Result<(), MessengerError>> =
            tokio::spawn(manage_conversation(config));
        handles.push(handle);
    }
    handles.shrink_to_fit();
    for handle in handles {
        handle.await??;
    }
    Ok(())
}
