// The state machine used to represent one end of a conversation.
// This includes inducing transitions and actions taken during transitions,
// so a lot of the messenger network code is here.

use mgen::{log, MessageHeader, SerializedMessage};
use rand_distr::Distribution;
use rand_xoshiro::Xoshiro256PlusPlus;
use std::borrow::Borrow;
use std::collections::HashMap;
use std::fmt::Debug;
use std::sync::Arc;
use tokio::sync::mpsc;
use tokio::time::Instant;

use crate::messenger::dists::Distributions;
use crate::messenger::message::{construct_message, construct_receipt};

/// All possible Conversation state machine states
pub enum StateMachine {
    Idle(Conversation<Idle>),
    Active(Conversation<Active>),
}

impl StateMachine {
    pub fn start(dists: Distributions, rng: &mut Xoshiro256PlusPlus) -> StateMachine {
        Self::Idle(Conversation::<Idle>::start(dists, rng))
    }
}

/// The state machine representing a conversation state and its transitions.
pub struct Conversation<S: State> {
    dists: Distributions,
    delay: Instant,
    state: S,
}

pub trait State {
    fn sent(conversation: Conversation<Self>, rng: &mut Xoshiro256PlusPlus) -> StateMachine
    where
        Self: Sized;
    fn received(conversation: Conversation<Self>, rng: &mut Xoshiro256PlusPlus) -> StateMachine
    where
        Self: Sized;
    fn to_machine(conversation: Conversation<Self>) -> StateMachine
    where
        Self: Sized;
}

pub struct Idle {}
pub struct Active {
    wait: Instant,
}

impl State for Idle {
    fn sent(conversation: Conversation<Self>, rng: &mut Xoshiro256PlusPlus) -> StateMachine {
        if conversation.dists.s.sample(rng) {
            log!("Idle: [sent] tranisition to [Active]");
            let delay = Instant::now() + conversation.dists.a_s.sample_secs(rng);
            let wait = Instant::now() + conversation.dists.w.sample_secs(rng);
            StateMachine::Active({
                Conversation::<Active> {
                    dists: conversation.dists,
                    delay,
                    state: Active { wait },
                }
            })
        } else {
            log!("Idle: [sent] tranisition to [Idle]");
            let delay = Instant::now() + conversation.dists.i.sample_secs(rng);
            StateMachine::Idle({
                Conversation::<Idle> {
                    dists: conversation.dists,
                    delay,
                    state: Idle {},
                }
            })
        }
    }

    fn received(conversation: Conversation<Self>, rng: &mut Xoshiro256PlusPlus) -> StateMachine {
        if conversation.dists.r.sample(rng) {
            log!("Idle: [recv'd] tranisition to [Active]");
            let wait = Instant::now() + conversation.dists.w.sample_secs(rng);
            let delay = Instant::now() + conversation.dists.a_r.sample_secs(rng);
            StateMachine::Active({
                Conversation::<Active> {
                    dists: conversation.dists,
                    delay,
                    state: Active { wait },
                }
            })
        } else {
            log!("Idle: [recv'd] tranisition to [Idle]");
            StateMachine::Idle(conversation)
        }
    }

    fn to_machine(conversation: Conversation<Self>) -> StateMachine {
        StateMachine::Idle(conversation)
    }
}

impl State for Active {
    fn sent(conversation: Conversation<Self>, rng: &mut Xoshiro256PlusPlus) -> StateMachine {
        log!("Active: [sent] transition to [Active]");
        let delay = Instant::now() + conversation.dists.a_s.sample_secs(rng);
        StateMachine::Active(Conversation::<Active> {
            dists: conversation.dists,
            delay,
            state: conversation.state,
        })
    }

    fn received(conversation: Conversation<Self>, rng: &mut Xoshiro256PlusPlus) -> StateMachine {
        log!("Active: [recv'd] transition to [Active]");
        let delay = Instant::now() + conversation.dists.a_r.sample_secs(rng);
        StateMachine::Active(Conversation::<Active> {
            dists: conversation.dists,
            delay,
            state: conversation.state,
        })
    }

    fn to_machine(conversation: Conversation<Self>) -> StateMachine {
        StateMachine::Active(conversation)
    }
}

impl Conversation<Idle> {
    fn start(dists: Distributions, rng: &mut Xoshiro256PlusPlus) -> Self {
        let delay = Instant::now() + dists.i.sample_secs(rng);
        log!("[start]");
        Self {
            dists,
            delay,
            state: Idle {},
        }
    }
}

impl Conversation<Active> {
    fn waited(self, rng: &mut Xoshiro256PlusPlus) -> Conversation<Idle> {
        log!("Active: [waited] tranision to [Idle]");
        let delay = Instant::now() + self.dists.i.sample_secs(rng);
        Conversation::<Idle> {
            dists: self.dists,
            delay,
            state: Idle {},
        }
    }

    async fn sleep(delay: Instant, wait: Instant) -> ActiveGroupActions {
        if delay < wait {
            log!("delaying for {:?}", delay - Instant::now());
            tokio::time::sleep_until(delay).await;
            ActiveGroupActions::Send
        } else {
            log!("waiting for {:?}", wait - Instant::now());
            tokio::time::sleep_until(wait).await;
            ActiveGroupActions::Idle
        }
    }
}

/// Type for getting messages from the reader thread in the state thread.
pub type StateFromReader = mpsc::UnboundedReceiver<MessageHeader>;
/// Type for sending messages from the state thread to the writer thread.
pub struct StateToWriter<S: MessageHolder> {
    pub channel: mpsc::UnboundedSender<S>,
}

pub trait MessageHolder: Borrow<SerializedMessage> + Debug {
    fn new(m: SerializedMessage) -> Self;
    fn clone(&self) -> Self;
}

impl MessageHolder for Arc<SerializedMessage> {
    fn new(m: SerializedMessage) -> Self {
        Self::new(m)
    }

    fn clone(&self) -> Self {
        Clone::clone(self)
    }
}

impl MessageHolder for Box<SerializedMessage> {
    fn new(m: SerializedMessage) -> Self {
        Self::new(m)
    }

    fn clone(&self) -> Self {
        panic!("Box holders should never clone");
    }
}

pub trait StreamMap<'a, S: 'a + MessageHolder, I: Iterator<Item = &'a mut StateToWriter<S>>> {
    fn channel_for(&self, name: &str) -> &StateToWriter<S>;
    fn values(&'a mut self) -> I;
}

impl<'a, S: MessageHolder>
    StreamMap<'a, S, std::collections::hash_map::ValuesMut<'a, String, StateToWriter<S>>>
    for HashMap<String, StateToWriter<S>>
{
    fn channel_for(&self, name: &str) -> &StateToWriter<S> {
        &self[name]
    }

    fn values(&'a mut self) -> std::collections::hash_map::ValuesMut<'a, String, StateToWriter<S>> {
        self.values_mut()
    }
}

impl<'a, S: MessageHolder> StreamMap<'a, S, std::iter::Once<&'a mut StateToWriter<S>>>
    for StateToWriter<S>
{
    fn channel_for(&self, _name: &str) -> &StateToWriter<S> {
        self
    }

    fn values(&'a mut self) -> std::iter::Once<&'a mut StateToWriter<S>> {
        std::iter::once(self)
    }
}

async fn send_action<
    'a,
    S: 'a + MessageHolder,
    T: State,
    I: ExactSizeIterator<Item = &'a mut StateToWriter<S>>,
>(
    conversation: Conversation<T>,
    mut streams: I,
    our_id: &str,
    group: &str,
    recipients: Vec<&str>,
    rng: &mut Xoshiro256PlusPlus,
) -> StateMachine {
    let size = conversation.dists.m.sample(rng);
    log!(
        "sending message from {} to {:?} of size {}",
        our_id,
        recipients,
        size
    );
    let m = S::new(construct_message(
        our_id.to_string(),
        group.to_string(),
        size,
    ));

    if streams.len() == 1 {
        streams
            .next()
            .unwrap()
            .channel
            .send(m)
            .expect("Internal stream closed with messages still being sent");
    } else {
        for stream in streams {
            stream
                .channel
                .send(m.clone())
                .expect("Internal stream closed with messages still being sent");
        }
    }

    T::sent(conversation, rng)
}

async fn receive_action<
    'a,
    S: 'a + MessageHolder,
    T: State,
    I: std::iter::Iterator<Item = &'a mut StateToWriter<S>>,
    M: StreamMap<'a, S, I>,
>(
    msg: MessageHeader,
    conversation: Conversation<T>,
    stream_map: &mut M,
    our_id: &str,
    rng: &mut Xoshiro256PlusPlus,
) -> StateMachine {
    match msg.body {
        mgen::MessageBody::Size(size) => {
            log!(
                "{} got message from {} of size {}",
                msg.group,
                msg.sender,
                size
            );
            let stream = stream_map.channel_for(&msg.sender);
            let m = construct_receipt(our_id.to_string(), msg.sender);
            stream
                .channel
                .send(S::new(m))
                .expect("channel from receive_action to sender closed");
            T::received(conversation, rng)
        }
        mgen::MessageBody::Receipt => T::to_machine(conversation),
    }
}

enum IdleGroupActions {
    Send,
    Receive(MessageHeader),
}

/// Handle a state transition from Idle, including I/O, for a multi-connection conversation.
/// Used for Idle group p2p conversations.
pub async fn manage_idle_conversation<
    'a,
    S: 'a + MessageHolder,
    I: std::iter::ExactSizeIterator<Item = &'a mut StateToWriter<S>>,
    M: StreamMap<'a, S, I> + 'a,
>(
    conversation: Conversation<Idle>,
    inbound: &mut StateFromReader,
    stream_map: &'a mut M,
    our_id: &str,
    group: &str,
    recipients: Vec<&str>,
    rng: &mut Xoshiro256PlusPlus,
) -> StateMachine {
    log!("delaying for {:?}", conversation.delay - Instant::now());
    let action = tokio::select! {
        () = tokio::time::sleep_until(conversation.delay) => IdleGroupActions::Send,

        res = inbound.recv() =>
            IdleGroupActions::Receive(res.expect("inbound channel closed")),
    };

    match action {
        IdleGroupActions::Send => {
            send_action(
                conversation,
                stream_map.values(),
                our_id,
                group,
                recipients,
                rng,
            )
            .await
        }
        IdleGroupActions::Receive(msg) => {
            receive_action(msg, conversation, stream_map, our_id, rng).await
        }
    }
}

enum ActiveGroupActions {
    Send,
    Receive(MessageHeader),
    Idle,
}

/// Handle a state transition from Active.
pub async fn manage_active_conversation<
    'a,
    S: 'a + MessageHolder,
    I: std::iter::ExactSizeIterator<Item = &'a mut StateToWriter<S>>,
    M: StreamMap<'a, S, I> + 'a,
>(
    conversation: Conversation<Active>,
    inbound: &mut StateFromReader,
    stream_map: &'a mut M,
    our_id: &str,
    group: &str,
    recipients: Vec<&str>,
    rng: &mut Xoshiro256PlusPlus,
) -> StateMachine {
    let action = tokio::select! {
        action = Conversation::<Active>::sleep(conversation.delay, conversation.state.wait) => action,

        res = inbound.recv() =>
            ActiveGroupActions::Receive(res.expect("inbound channel closed")),
    };

    match action {
        ActiveGroupActions::Send => {
            send_action(
                conversation,
                stream_map.values(),
                our_id,
                group,
                recipients,
                rng,
            )
            .await
        }
        ActiveGroupActions::Receive(msg) => {
            receive_action(msg, conversation, stream_map, our_id, rng).await
        }
        ActiveGroupActions::Idle => StateMachine::Idle(conversation.waited(rng)),
    }
}
